@@ -121,9 +121,9 @@ let%expect_test "checking setters (should end in [_exn])" =
   let test setters z =
     try_setters z setters
     |> List.iteri ~f:(fun i -> function
-      | Ok bytes ->
-        raise_s [%message "didn't raise" (z : Int.Hex.t) (i : int) (bytes : int list)]
-      | Error _ -> ())
+         | Ok bytes ->
+           raise_s [%message "didn't raise" (z : Int.Hex.t) (i : int) (bytes : int list)]
+         | Error _ -> ())
   in
   test [ set_int8_exn ] 0x80;
   test [ set_uint8_exn ] (-1);
@@ -155,6 +155,8 @@ module Local = struct
   let get_int64_t_be = get_int64_t_be
   let unsafe_get_int64_t_le = unsafe_get_int64_t_le
   let unsafe_get_int64_t_be = unsafe_get_int64_t_be
+  let get_string = get_string
+  let unsafe_get_string = unsafe_get_string
 end
 
 let%test_module "truncating setters (should end in [_trunc] or begin with [unsafe_])" =
@@ -162,8 +164,8 @@ let%test_module "truncating setters (should end in [_trunc] or begin with [unsaf
     let test setters z =
       try_setters z setters
       |> List.iter ~f:(fun t ->
-        Or_error.ok_exn t |> List.iter ~f:(printf "%x ");
-        printf "; ")
+           Or_error.ok_exn t |> List.iter ~f:(printf "%x ");
+           printf "; ")
     ;;
 
     let%expect_test "all word sizes" =
@@ -211,75 +213,91 @@ let%test_module "truncating getters (should end in [_trunc] or begin with [unsaf
         ; 0xc1 (* negative if top bit truncated *)
         ]
         ~f:(fun first_byte ->
-          let i = getter (getter_t ~first_byte) ~pos:0 in
-          (* Signed hex is not clear; make sure the hex is unsigned.  Include the signed
+        let i = getter (getter_t ~first_byte) ~pos:0 in
+        (* Signed hex is not clear; make sure the hex is unsigned.  Include the signed
              decimal form mainly to indicate the sign. *)
-          printf !"0x%x (= %d)\n" i i)
+        printf !"0x%x (= %d)\n" i i)
     ;;
 
     let%expect_test ("63-bit int" [@tags "64-bits-only"]) =
       test get_int64_le_trunc;
       [%expect
         {|
-          0x887868584838281 (= 614607782171345537)
-          0x48c7c6c5c4c3c2c1 (= -3978993193046523199) |}];
+        0x887868584838281 (= 614607782171345537)
+        0x48c7c6c5c4c3c2c1 (= -3978993193046523199)
+        |}];
       test get_int64_be_trunc;
       [%expect
         {|
-          0x182838485868788 (= 108793946209421192)
-          0x41c2c3c4c5c6c7c8 (= -4484807029008447544) |}];
+        0x182838485868788 (= 108793946209421192)
+        0x41c2c3c4c5c6c7c8 (= -4484807029008447544)
+        |}];
       test unsafe_get_int64_le_trunc;
       [%expect
         {|
-          0x887868584838281 (= 614607782171345537)
-          0x48c7c6c5c4c3c2c1 (= -3978993193046523199) |}];
+        0x887868584838281 (= 614607782171345537)
+        0x48c7c6c5c4c3c2c1 (= -3978993193046523199)
+        |}];
       test unsafe_get_int64_be_trunc;
       [%expect
         {|
-          0x182838485868788 (= 108793946209421192)
-          0x41c2c3c4c5c6c7c8 (= -4484807029008447544) |}]
+        0x182838485868788 (= 108793946209421192)
+        0x41c2c3c4c5c6c7c8 (= -4484807029008447544)
+        |}]
     ;;
 
     let%expect_test ("31-bit int" [@tags "32-bits-only", "no-js"]) =
       test get_int64_le_trunc;
-      [%expect {|
-          0x4838281 (= 75727489)
-          0x44c3c2c1 (= -993803583) |}];
+      [%expect
+        {|
+        0x4838281 (= 75727489)
+        0x44c3c2c1 (= -993803583)
+        |}];
       test get_int64_be_trunc;
-      [%expect {|
-          0x5868788 (= 92702600)
-          0x45c6c7c8 (= -976828472) |}];
+      [%expect
+        {|
+        0x5868788 (= 92702600)
+        0x45c6c7c8 (= -976828472)
+        |}];
       test unsafe_get_int64_le_trunc;
-      [%expect {|
-          0x4838281 (= 75727489)
-          0x44c3c2c1 (= -993803583) |}];
+      [%expect
+        {|
+        0x4838281 (= 75727489)
+        0x44c3c2c1 (= -993803583)
+        |}];
       test unsafe_get_int64_be_trunc;
-      [%expect {|
-          0x5868788 (= 92702600)
-          0x45c6c7c8 (= -976828472) |}]
+      [%expect
+        {|
+        0x5868788 (= 92702600)
+        0x45c6c7c8 (= -976828472)
+        |}]
     ;;
 
     let%expect_test ("32-bit int" [@tags "js-only"]) =
       test get_int64_le_trunc;
       [%expect
         {|
-          0x84838281 (= -2071756159)
-          0xc4c3c2c1 (= -993803583) |}];
+        0x84838281 (= -2071756159)
+        0xc4c3c2c1 (= -993803583)
+        |}];
       test get_int64_be_trunc;
       [%expect
         {|
-          0x85868788 (= -2054781048)
-          0xc5c6c7c8 (= -976828472) |}];
+        0x85868788 (= -2054781048)
+        0xc5c6c7c8 (= -976828472)
+        |}];
       test unsafe_get_int64_le_trunc;
       [%expect
         {|
-          0x84838281 (= -2071756159)
-          0xc4c3c2c1 (= -993803583) |}];
+        0x84838281 (= -2071756159)
+        0xc4c3c2c1 (= -993803583)
+        |}];
       test unsafe_get_int64_be_trunc;
       [%expect
         {|
-          0x85868788 (= -2054781048)
-          0xc5c6c7c8 (= -976828472) |}]
+        0x85868788 (= -2054781048)
+        0xc5c6c7c8 (= -976828472)
+        |}]
     ;;
   end)
 ;;
@@ -303,8 +321,8 @@ let%expect_test "checking getters (should end in [_exn])" =
   let test getters ~first_bigstring_byte =
     try_getters getters ~first_bigstring_byte
     |> List.iteri ~f:(fun i -> function
-      | Ok z -> raise_s [%message "didn't raise" (i : int) (z : Int.Hex.t)]
-      | Error _ -> ())
+         | Ok z -> raise_s [%message "didn't raise" (i : int) (z : Int.Hex.t)]
+         | Error _ -> ())
   in
   test
     (* These should check that the 64th bit in the string representation is redundant, so
@@ -357,8 +375,9 @@ let%expect_test "basic int64 setters" =
   |> printf !"%{sexp#hum:int list Or_error.t list}\n";
   [%expect
     {|
-      ((Ok (1 2 3 4 5 6 7 8)) (Ok (8 7 6 5 4 3 2 1)) (Ok (1 2 3 4 5 6 7 8))
-       (Ok (8 7 6 5 4 3 2 1))) |}]
+    ((Ok (1 2 3 4 5 6 7 8)) (Ok (8 7 6 5 4 3 2 1)) (Ok (1 2 3 4 5 6 7 8))
+     (Ok (8 7 6 5 4 3 2 1)))
+    |}]
 ;;
 
 let unsafe_get_int64_t_be = unsafe_get_int64_t_be
@@ -381,9 +400,10 @@ let%expect_test "basic int64 getters" =
   |> printf !"%{sexp#hum:Int64.Hex.t Or_error.t list}\n";
   [%expect
     {|
-      ((Ok 0x102030405060708) (Ok 0x807060504030201) (Ok 0x102030405060708)
-       (Ok 0x807060504030201) (Ok 0x102030405060708) (Ok 0x807060504030201)
-       (Ok 0x102030405060708) (Ok 0x807060504030201)) |}]
+    ((Ok 0x102030405060708) (Ok 0x807060504030201) (Ok 0x102030405060708)
+     (Ok 0x807060504030201) (Ok 0x102030405060708) (Ok 0x807060504030201)
+     (Ok 0x102030405060708) (Ok 0x807060504030201))
+    |}]
 ;;
 
 let unsafe_set_int32_t_be = unsafe_set_int32_t_be
@@ -398,8 +418,9 @@ let%expect_test "basic int32 setters" =
   |> printf !"%{sexp#hum:int list Or_error.t list}\n";
   [%expect
     {|
-      ((Ok (1 2 3 4 0 0 0 0)) (Ok (4 3 2 1 0 0 0 0)) (Ok (1 2 3 4 0 0 0 0))
-       (Ok (4 3 2 1 0 0 0 0))) |}]
+    ((Ok (1 2 3 4 0 0 0 0)) (Ok (4 3 2 1 0 0 0 0)) (Ok (1 2 3 4 0 0 0 0))
+     (Ok (4 3 2 1 0 0 0 0)))
+    |}]
 ;;
 
 let unsafe_get_int32_t_be = unsafe_get_int32_t_be
@@ -427,8 +448,9 @@ let%expect_test "basic int setters" =
   |> printf !"%{sexp#hum:int list Or_error.t list}\n";
   [%expect
     {|
-      ((Ok (0 0 0 0 1 2 3 4)) (Ok (4 3 2 1 0 0 0 0)) (Ok (0 0 0 0 1 2 3 4))
-       (Ok (4 3 2 1 0 0 0 0))) |}]
+    ((Ok (0 0 0 0 1 2 3 4)) (Ok (4 3 2 1 0 0 0 0)) (Ok (0 0 0 0 1 2 3 4))
+     (Ok (4 3 2 1 0 0 0 0)))
+    |}]
 ;;
 
 let unsafe_get_uint32_be = unsafe_get_uint32_be
@@ -479,14 +501,42 @@ let%expect_test "basic int getters" =
   |> printf !"%{sexp#hum:Int.Hex.t Or_error.t list}\n";
   [%expect
     {|
-      ((Ok 0x1020304) (Ok 0x4030201) (Ok 0x1020304) (Ok 0x4030201) (Ok 0x1020304)
-       (Ok 0x4030201) (Ok 0x1020304) (Ok 0x4030201) (Ok 0x102) (Ok 0x201)
-       (Ok 0x102) (Ok 0x201) (Ok 0x102) (Ok 0x201) (Ok 0x102) (Ok 0x201) (Ok 0x1)
-       (Ok 0x1) (Ok 0x1) (Ok 0x1)) |}]
+    ((Ok 0x1020304) (Ok 0x4030201) (Ok 0x1020304) (Ok 0x4030201) (Ok 0x1020304)
+     (Ok 0x4030201) (Ok 0x1020304) (Ok 0x4030201) (Ok 0x102) (Ok 0x201)
+     (Ok 0x102) (Ok 0x201) (Ok 0x102) (Ok 0x201) (Ok 0x102) (Ok 0x201) (Ok 0x1)
+     (Ok 0x1) (Ok 0x1) (Ok 0x1))
+    |}]
+;;
+
+let get_string = get_string
+let unsafe_get_string = unsafe_get_string
+
+let%expect_test "basic string getters" =
+  for len = 0 to 4 do
+    try_getters
+      ~first_bigstring_byte:1
+      [ get_string ~len
+      ; unsafe_get_string ~len
+      ; (fun t ~pos -> (Local.get_string t ~pos ~len |> globalize_string) [@nontail])
+      ; (fun t ~pos ->
+          (Local.unsafe_get_string t ~pos ~len |> globalize_string) [@nontail])
+      ]
+    |> printf !"%{sexp#hum:string Or_error.t list}\n"
+  done;
+  [%expect
+    {|
+    ((Ok "") (Ok "") (Ok "") (Ok ""))
+    ((Ok "\001") (Ok "\001") (Ok "\001") (Ok "\001"))
+    ((Ok "\001\002") (Ok "\001\002") (Ok "\001\002") (Ok "\001\002"))
+    ((Ok "\001\002\003") (Ok "\001\002\003") (Ok "\001\002\003")
+     (Ok "\001\002\003"))
+    ((Ok "\001\002\003\004") (Ok "\001\002\003\004") (Ok "\001\002\003\004")
+     (Ok "\001\002\003\004"))
+    |}]
 ;;
 
 external unsafe_find : t_frozen -> char -> pos:int -> len:int -> int = "bigstring_find"
-[@@noalloc]
+  [@@noalloc]
 
 let%expect_test "basic unsafe_find" =
   let t = of_string "abc" in
@@ -523,7 +573,7 @@ external unsafe_memmem
   -> needle_len:int
   -> int
   = "bigstring_memmem_bytecode" "bigstring_memmem"
-[@@noalloc]
+  [@@noalloc]
 
 let%expect_test "basic unsafe_memmem" =
   let haystack = "foo bar baz qwux" |> of_string in
@@ -596,6 +646,7 @@ let%expect_test "basic get_opt_len" =
 
 let memcmp = memcmp
 let memcmp_bytes = memcmp_bytes
+let memcmp_string = memcmp_string
 
 let%test_module "basic memcmp" =
   (module struct
@@ -615,6 +666,10 @@ let%test_module "basic memcmp" =
     let%expect_test "bigstring to bytes" =
       test_memcmp ~memcmp:memcmp_bytes (of_string s1) (Bytes.of_string s2)
     ;;
+
+    let%expect_test "bigstring to string" =
+      test_memcmp ~memcmp:memcmp_string (of_string s1) s2
+    ;;
   end)
 ;;
 
@@ -629,21 +684,29 @@ let%expect_test "basic char setters" =
   |> printf !"%{sexp#hum:int list Or_error.t list}\n";
   [%expect
     {|
-      ((Ok (0 0 0 0 0 0 0 0)) (Ok (120 0 0 0 0 0 0 0)) (Ok (120 120 0 0 0 0 0 0))
-       (Ok (120 0 0 0 0 0 0 0))) |}]
+    ((Ok (0 0 0 0 0 0 0 0)) (Ok (120 0 0 0 0 0 0 0)) (Ok (120 120 0 0 0 0 0 0))
+     (Ok (120 0 0 0 0 0 0 0)))
+    |}]
 ;;
 
+let unsafe_memset = unsafe_memset
+
 external unsafe_set : t_frozen -> int -> char -> unit = "%caml_ba_unsafe_set_1"
 
 let%expect_test "basic char unsafe setters" =
   try_setters
     'x'
-    [ memset ~len:0; memset ~len:1; memset ~len:2; (fun t ~pos -> set t pos) ]
+    [ unsafe_memset ~len:0
+    ; unsafe_memset ~len:1
+    ; unsafe_memset ~len:2
+    ; (fun t ~pos -> unsafe_set t pos)
+    ]
   |> printf !"%{sexp#hum:int list Or_error.t list}\n";
   [%expect
     {|
-      ((Ok (0 0 0 0 0 0 0 0)) (Ok (120 0 0 0 0 0 0 0)) (Ok (120 120 0 0 0 0 0 0))
-       (Ok (120 0 0 0 0 0 0 0))) |}]
+    ((Ok (0 0 0 0 0 0 0 0)) (Ok (120 0 0 0 0 0 0 0)) (Ok (120 120 0 0 0 0 0 0))
+     (Ok (120 0 0 0 0 0 0 0)))
+    |}]
 ;;
 
 external get : t_frozen -> int -> char = "%caml_ba_ref_1"
@@ -691,18 +754,19 @@ let%expect_test "basic concat" =
           print_s [%message (sep : t option) (ts : t list) ~concat:(concat ?sep ts : t)]));
   [%expect
     {|
-      ((sep ())
-       (ts  ())
-       (concat ""))
-      ((sep ()) (ts (fst1)) (concat fst1))
-      ((sep ()) (ts (fst snd)) (concat fstsnd))
-      ((sep ()) (ts (fst3 snd3 trd)) (concat fst3snd3trd))
-      ((sep ("; ")) (ts ()) (concat ""))
-      ((sep ("; "))
-       (ts  (fst1))
-       (concat fst1))
-      ((sep ("; ")) (ts (fst snd)) (concat "fst; snd"))
-      ((sep ("; ")) (ts (fst3 snd3 trd)) (concat "fst3; snd3; trd")) |}]
+    ((sep ())
+     (ts  ())
+     (concat ""))
+    ((sep ()) (ts (fst1)) (concat fst1))
+    ((sep ()) (ts (fst snd)) (concat fstsnd))
+    ((sep ()) (ts (fst3 snd3 trd)) (concat fst3snd3trd))
+    ((sep ("; ")) (ts ()) (concat ""))
+    ((sep ("; "))
+     (ts  (fst1))
+     (concat fst1))
+    ((sep ("; ")) (ts (fst snd)) (concat "fst; snd"))
+    ((sep ("; ")) (ts (fst3 snd3 trd)) (concat "fst3; snd3; trd"))
+    |}]
 ;;
 
 let equal = equal
@@ -743,6 +807,138 @@ let%expect_test ("local allocation does not heap allocate" [@tags "64-bits-only"
   [%expect ""]
 ;;
 
+let%expect_test "unsafe_get_int64_le_exn correctness" =
+  let case (here, behavior, i) =
+    let convert () =
+      let t = create 8 in
+      set_int64_t_le t ~pos:0 i;
+      let result = unsafe_get_int64_le_exn t ~pos:0 in
+      let expected = Int.of_int64_trunc i in
+      require_equal here (module Int) expected result;
+      result
+    in
+    match behavior with
+    | `Fit -> require_does_not_raise here (fun () -> ignore (convert () : int))
+    | `Raise ->
+      Or_error.try_with convert
+      |> require_error here (fun result ->
+           [%message "Unexpectedly successfully converted" (result : int)])
+  in
+  let max_val = Int64.of_int Int.max_value in
+  let min_val = Int64.of_int Int.min_value in
+  List.iter
+    ~f:case
+    [ [%here], `Fit, 0L
+    ; [%here], `Fit, 1000000L
+    ; [%here], `Fit, -1000000L
+    ; [%here], `Fit, max_val
+    ; ([%here], `Fit, Int64.(max_val - 1L))
+    ; ([%here], `Fit, Int64.(max_val - 1000L))
+    ; ([%here], `Raise, Int64.(max_val + 1L))
+    ; ([%here], `Raise, Int64.(max_val + 1000L))
+    ; [%here], `Fit, min_val
+    ; ([%here], `Fit, Int64.(min_val + 1L))
+    ; ([%here], `Fit, Int64.(min_val + 1000L))
+    ; ([%here], `Raise, Int64.(min_val - 1L))
+    ; ([%here], `Raise, Int64.(min_val - 1000L))
+    ]
+;;
+
+let%bench_module "" =
+  (module struct
+    (* Copied some stuff so we could benchmark int64_to_int_exn different implementations.
+    *)
+    let arch_sixtyfour = Stdlib.Sys.word_size = 64
+
+    external int64_to_int : int64 -> int = "%int64_to_int"
+
+    let int64_conv_error () =
+      failwith "unsafe_read_int64: value cannot be represented unboxed!"
+    ;;
+
+    let some_int = 42L
+
+    (* [Poly] is required so that we can compare unboxed [int64]. *)
+    let[@inline always] old_int64_to_int_exn n =
+      if arch_sixtyfour
+      then
+        if Poly.(n >= -0x4000_0000_0000_0000L && n < 0x4000_0000_0000_0000L)
+        then int64_to_int n
+        else int64_conv_error ()
+      else if Poly.(n >= -0x0000_0000_4000_0000L && n < 0x0000_0000_4000_0000L)
+      then int64_to_int n
+      else int64_conv_error ()
+    ;;
+
+    let[@inline always] bit_manipulation_int64_to_int_exn n =
+      if arch_sixtyfour
+      then
+        (*
+           For positive int64, the bits must start with: 00...
+           and for negative ones, the bits must start with: 11...
+           {v
+               n           = 0bXY...
+               m = n asr 1 = 0bXX...
+               m lxor n    = 0b0(X xor Y)...
+                           = 0b01... if n = 0b01... or 0b10...
+                           = 0b00... if n = 0b00... or 0b11...
+             v}
+        *)
+        if Poly.(Int64.((n asr 1) lxor n) < 0x4000_0000_0000_0000L)
+        then int64_to_int n
+        else int64_conv_error ()
+      else if Poly.(n >= -0x0000_0000_4000_0000L && n < 0x0000_0000_4000_0000L)
+      then int64_to_int n
+      else int64_conv_error ()
+    ;;
+
+    let[@inline always] with_poly_eq_int64_to_int_exn n =
+      let n' = int64_to_int n in
+      if Poly.( = ) (Int64.of_int n') n then n' else int64_conv_error ()
+    ;;
+
+    let%bench_fun "with_poly_eq (new implementation)" =
+      fun () ->
+      for _ = 1 to 1000 do
+        ignore
+          (Sys.opaque_identity
+             (with_poly_eq_int64_to_int_exn (Sys.opaque_identity some_int))
+            : int)
+      done
+    ;;
+
+    let%bench_fun "bit manipulation" =
+      fun () ->
+      for _ = 1 to 1000 do
+        ignore
+          (Sys.opaque_identity
+             (bit_manipulation_int64_to_int_exn (Sys.opaque_identity some_int))
+            : int)
+      done
+    ;;
+
+    let%bench_fun "old implementation (doing range checks)" =
+      fun () ->
+      for _ = 1 to 1000 do
+        ignore
+          (Sys.opaque_identity (old_int64_to_int_exn (Sys.opaque_identity some_int))
+            : int)
+      done
+    ;;
+
+    (*
+       Results:
+       ┌──────────────────────────────────────────────────────────────┬────────────┬────────────┐
+       │ Name                                                         │   Time/Run │ Percentage │
+       ├──────────────────────────────────────────────────────────────┼────────────┼────────────┤
+       │ [test_bigstring.ml:] with_poly_eq (new implementation)       │   579.22ns │     39.53% │
+       │ [test_bigstring.ml:] bit manipulation                        │   985.03ns │     67.22% │
+       │ [test_bigstring.ml:] old implementation (doing range checks) │ 1_465.38ns │    100.00% │
+       └──────────────────────────────────────────────────────────────┴────────────┴────────────┘
+    *)
+  end)
+;;
+
 type nonrec t = t
 type nonrec t_frozen = t_frozen [@@deriving compare, hash, sexp]
 
